// Parameter class for medical sequence segmentation using graph cut

#ifndef CLASSE_PARAMETRES_SEGVID
#define CLASSE_PARAMETRES_SEGVID

#include "various_tools.h"

namespace MRFSegm {

  struct Params {
    
    // Name of the parameter file
    string nom_fichier_param;
    
    // see details in reading function
    int version;
    string input_prefix;
    int nb_seq; 
    int nt;
    int t0; 

    string* input_prob_files;
    string* input_inter_files;

    int H;
    int W;
    int D;

    double zoom_factor;

    string proba_type;
    string data_type;

    list< pair<int,int> > Inclusions;

    double interaction_coeff;
    double inverse_sigma_fact;

    bool verbose_base;

    string output_path;
    bool save_volume;
    string output_syntax_volume;
    bool save_slice;
    string output_syntax_slice;


  Params(){
    input_prob_files = NULL;
    input_inter_files = NULL;
  }

  ~Params(){
    if (input_prob_files != NULL)
      delete[] input_prob_files;
    if (input_inter_files != NULL)
      delete[] input_inter_files;
  }
	
  void charge_params(int nb_arg, char** liste_arg){

    string nom_fic_param;
    if (nb_arg != 2){
      cout << "No parameter file specified. Using ./params.dat..." << endl;
      nom_fic_param = "params.dat";
    } else {
      stringstream sss; sss << liste_arg[1];  sss >> nom_fic_param;
    }
    charge_params(nom_fic_param);
  }


  void charge_params(string nom_fic_param){

    // Opening parameter file
    cout << "Non interactive mode. Parameter file : " << nom_fic_param << endl << endl;
    nom_fichier_param = nom_fic_param;
    ifstream fp(nom_fic_param.c_str());

    // Reading parameters
    lire_param(fp, "Version of parameter file", version);

    // Data
    lire_param(fp, "General input prefix", input_prefix);
    lire_param(fp, "Nb of sequences", nb_seq);
    lire_param(fp, "Number of frames", nt);
    lire_param(fp, "Initial frame", t0);

    cout << "Label probabilities to be read in :" << endl;
    input_prob_files = new string[nb_seq];
    lire_liste_param(fp, "`- file name for seq", input_prob_files, nb_seq);

    cout << "Intensity values, to estimate similarity between neighbors :" << endl;
    input_inter_files = new string[nb_seq];
    lire_liste_param(fp, "`- file name for seq", input_inter_files, nb_seq);

    cout << "Original image size :" << endl;
    lire_param(fp, "Height", H);
    lire_param(fp, "Width", W);
    lire_param(fp, "Depth", D);

    lire_param(fp, "Sampling factor", zoom_factor);

    lire_param(fp, "Proba type", proba_type);
    lire_param(fp, "Data type", data_type);

    // Inclusions
    int nb_incl;
    lire_param(fp, "Number of inclusion constraints", nb_incl);
    for (int i = 0; i<nb_incl; i++){
      int a,b;
      //      cout << "Inclusion " << i+1 << " :" << endl;
      lire_param(fp, " Foreground in sequence", a);
      lire_param(fp, "  is included in the foreground of sequence", b);
      Inclusions.push_back( pair<int,int>(a,b) );
    }

    // Coefficients
    lire_param(fp, "Interaction relative coefficient", interaction_coeff);
    lire_param(fp, "Inverse standard-deviation factor", inverse_sigma_fact);

    // Various
    lire_param(fp, "Verbose", verbose_base);

    // Outputs
    lire_param(fp, "Output path", output_path);

    if (version > 1){
      lire_param(fp, "Save volume ?", save_volume);
      lire_param(fp, "Output syntax for volume", output_syntax_volume);
      lire_param(fp, "Save slices ?", save_slice);
    } else {
      save_volume = false;
      output_syntax_volume = "%sseq%d_time%d.dat";
      save_slice = true;
      cout << BoldGreen << "Parameters by default for this old version 1 :" << BoldRed << " !!!!!!!!!!!!!!!!!!!" << endl;
      cout << BoldGreen << "  save volume : " << TextCyan << "no" << endl;
      cout << BoldGreen << "  Output syntax for volume : " << TextCyan << "%sseq%d_time%d.dat" << endl;
      cout << BoldGreen << "  save slices : " << TextCyan << "yes" << TextNormal << endl;
    }
    lire_param(fp, "Output syntax for slices", output_syntax_slice);

    
    // File must end with "ok"
    string v_aux;
    lire_param(fp, "End of parameter file... ", v_aux);
    if (!(fp.good() && (v_aux=="ok")))
      cout << '\a' << endl << "------ ERROR end of parameter file ------" << endl;
    cout << endl;
    fp.close();
  }
	
};


// Copy the parameter file in the result directory
void sauver_params(int nb_arg, char** liste_arg, string output){
  if (nb_arg != 2)
    cout << '\a' << endl << "On ne peut pas recopier le fichier de paramtres. Raison: pb argument ligne de commande, il faut prciser UN nom de fichier de paramtres." << endl;
  else {
    string nom_fic_param;
    stringstream sss; sss << liste_arg[1];  sss >> nom_fic_param;
    char* s_aux = new char[300];
    sprintf(s_aux, "cp -i %s %s", nom_fic_param.c_str(), output.c_str());
    system(s_aux);
    delete[] s_aux;
  }
}

//idem without useless considerations
void sauver_params(string nom_fic_param, string output){
  char* s_aux = new char[300];
  sprintf(s_aux, "cp -i %s %s", nom_fic_param.c_str(), output.c_str());
  system(s_aux);
  delete[] s_aux;
}

} // end of namespace

#endif
